{
Copyright (c) 2012 Yvon Massé
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}

unit Unit1;

{$MODE Delphi}

interface

uses
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Math, LCLProc,
  {$IFDEF WIN32}
    Windows, ShellApi;
  {$ENDIF}
  {$IFDEF LINUX}
    Process;
  {$ENDIF}

type

  { TForm1 }

  TForm1 = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Conf: TComboBox;
    ChkPa: TCheckBox;
    ChkPb: TCheckBox;
    ChkPc: TCheckBox;
    ChkGa: TCheckBox;
    ChkGb: TCheckBox;
    ChkGc: TCheckBox;
    EdPa: TEdit;
    EdPb: TEdit;
    EdPc: TEdit;
    EdGa: TEdit;
    EdGb: TEdit;
    EdGc: TEdit;
    UniPa: TComboBox;
    UniPb: TComboBox;
    UniPc: TComboBox;
    UniGa: TComboBox;
    UniGb: TComboBox;
    UniGc: TComboBox;
    BtCalcul: TButton;
    BtRaz: TButton;
    BtConfig: TButton;
    BtLisez: TButton;
    Info: TMemo;
    procedure ConfChange(Sender: TObject);
    procedure ChkPaClick(Sender: TObject);
    procedure ChkPbClick(Sender: TObject);
    procedure ChkPcClick(Sender: TObject);
    procedure ChkGaClick(Sender: TObject);
    procedure ChkGbClick(Sender: TObject);
    procedure ChkGcClick(Sender: TObject);
    procedure EdPaExit(Sender: TObject);
    procedure EdPbExit(Sender: TObject);
    procedure EdPcExit(Sender: TObject);
    procedure EdGaExit(Sender: TObject);
    procedure EdGbExit(Sender: TObject);
    procedure EdGcExit(Sender: TObject);
    procedure UniPaChange(Sender: TObject);
    procedure UniPbChange(Sender: TObject);
    procedure UniPcChange(Sender: TObject);
    procedure UniGaChange(Sender: TObject);
    procedure UniGbChange(Sender: TObject);
    procedure UniGcChange(Sender: TObject);
    procedure BtCalculClick(Sender: TObject);
    procedure BtRazClick(Sender: TObject);
    procedure BtConfigClick(Sender: TObject);
    procedure BtLisezClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
  private
    { Déclarations privées }
  public
    { Déclarations publiques }
  end;

const
prc = 6;  // Nombre de chiffre des résultats

var
  Form1: TForm1;
  LiC: Tlist;                         // Liste des objets case à cocher
  LiE: Tlist;                         // Liste des objets champ de saisie
  LiB: Tlist;                         // Liste des objets boîte de sélection
  vcf: Integer = 0;                   // Version du fichier de configuration
  Tbol: Array[0..5] of Boolean;       // Table des valeurs d'entrée valides
  Tvar: Array[0..5, 0..1] of Real;    // Tables des valeurs angulaires en °
  Etiq: Array[0..5] of String;        // Tables des mnémoniques
  Comp: Array[0..5] of Boolean;       // Tables des complémentations
  Srot: Array[3..5] of Integer;       // Tables des signes
  Grot: Integer;                      // Contexte de travail
  Unt: Array[0..5] of String;         // Table des unités

implementation

{$R *.lfm}

function vals(ch, sp: String; var vl: Real): Boolean;
// Evalue la chaîne ch pouvant contenir des séparateurs sp pour le système sexagésimal
var
  i, e, b: Integer;
  v: Real;
begin
  vl := 0; b := 1;
  Result := False;
  if Length(ch) = 0 then
    Result := True
  else
    repeat
      i := Pos(sp, ch); // Recherche de sp dans ch
      if i = 0 then
        Val(ch, v, e)   // sp non trouvé, évaluation de ch
      else
      begin
        Val(Copy(ch, 1, i - 1), v, e); // Evaluation du chiffre avant sp
        if i = UTF8Length(ch) then
          i := 0            // L'évaluation est terminée
        else
          UTF8Delete(ch, 1, i); // Sinon on supprime la partie évaluée
      end;
      if e = 0 then
      begin                 // Pas d'erreur, on prend en compte
        vl := vl + v/b;
        b := 60*b; if v < 0 then b := -b;
      end
      else
        Result := True;
    until Result or (i = 0);
end;

function lecs(ch: String; i: Integer; var vl: Real): Boolean;
// Lecture de la saisie d'indice i
var
  e: Integer;
begin
  if Unt[i] = '°' then
    Result := vals(ch, '°', vl)
  else
    if UpperCase(Unt[i])[1] = 'H' then
    begin
      Result := vals(ch, ':', vl);
      vl := vl*15;
      if Length(Unt[i]) <> 1 then     // Cas des heures vraies (hv)
      begin
        vl := vl - 180;
        if (vl < 0) and (Abs(Srot[i]) = 2) then vl := vl + 360;
      end;
    end
    else
    begin
      Val(ch, vl, e);
      Result := e <> 0;
      case UpperCase(Unt[i])[1] of
      'R': vl := vl*180/Pi;
      'G': vl := vl*0.9;
      'K': vl := vl*0.009;
      'M': vl := vl/60;
      end;
    end;
end;

function mods(i: Integer): Boolean;
var
  bl: Boolean;
begin
  if i >= 3 then
  begin
    if Grot <> 0 then
    begin
      // Contexte de travail : application pratique
      if Tbol[3 + (i - 2) mod 3] or Tbol[3 + (i - 1) mod 3] then
      begin
      // Un autre angle a déjà été saisi, vérification de la compatibilité
        case Grot*Srot[i] of
        1, -1: begin
          Tvar[i, 0] := Grot*Srot[i]*Tvar[i, 0];
          bl := Tvar[i, 0] < 0;
          end;
        2, -2: begin
          if Grot*Srot[i] < 0 then Tvar[i, 0] := 360 - Tvar[i, 0];
          bl := Tvar[i, 0] > 180;
          end;
        end;
        if bl then
          MessageDlgPos('Incompatibilité des valeurs d''angle', mtError, [mbOK], 0, Form1.Left, Form1.Top);
      end
      else
      begin
      // Premier angle saisi, il détermine la valeur de Grot
        case Srot[i] of
        1, -1:
          if Tvar[i, 0] < 0 then
          begin
            Tvar[i, 0] := -Tvar[i, 0];
            Grot := -Srot[i];
          end
          else
            Grot := Srot[i];
        2, -2:
          if Tvar[i, 0] > 180 then
          begin
            Tvar[i, 0] := 360 - Tvar[i, 0];
            Grot := -Srot[i] div 2;
          end
          else
            Grot := Srot[i] div 2;
        end;
      end;
    end;
    if Comp[i] then Tvar[i, 0] := 180 - Tvar[i, 0];
  end
  else
    if Comp[i] then Tvar[i, 0] := 90 - Tvar[i, 0];
  Result := (Tvar[i, 0] <= 0) or (Tvar[i, 0] >= 180);
end;

function modr(i: Integer): Real;
// Complémentation et modification de la valeur de l'angle en fonction de Grot
begin
    Result := Tvar[i, 0];
  if Comp[i] then                    // Complémentation ?
    if i < 3 then
      Result := 90 - Result
    else
      Result := 180 - Result;
  if (i >= 3) and (Grot <> 0) then   // Valeur de l'angle en fonction de Grot
    case Grot*Srot[i] of
    -1: Result := -Result;
    -2: Result := 360 - Result;
    end;
end;

function affr(rst: Real; i, pr: Integer): String;
// Affichage du résultat d'indice i
begin
  if Unt[i] = '°' then
    Result := FloatToStrF(rst, ffGeneral, pr, 1)
  else
    case UpperCase(Unt[i])[1] of
    'H':
      if Length(Unt[i]) = 1 then
      begin
        Result := TimeToStr(Abs(rst/360));
        if rst < 0 then Result := '-' + Result;
      end
      else              // Cas des heures vraies (hv)
        Result := TimeToStr(rst/360 + 0.5);
    'R': Result := FloatToStrF(rst*Pi/180, ffGeneral, pr, 1);
    'G': Result := FloatToStrF(rst/0.9, ffGeneral, pr, 1);
    'K': Result := FloatToStrF(rst/0.009, ffGeneral, pr, 1);
    'M': Result := FloatToStrF(rst*60, ffGeneral, pr, 1);
    end
end;

function chnt(i: Integer): String;
// Retourne le texte de l'info-bulle en fonction de Unt[i]
begin
  if Unt[i] = '°' then
    Result := 'Format : [-]dd°mm°ss'
  else
    case UpperCase(Unt[i])[1] of
    'H':
      if Length(Unt[i]) = 1 then
        Result := 'Format : [-]hh:mm:ss'
      else
        Result := 'Format : hh:mm:ss';
    else Result := '';
    end;
end;

Procedure ChkClick(i: Integer);
// Routine après clic sur boîte à cocher d'indice i
begin
  if TCheckBox(LiC[i]).Checked then
  begin
    TEdit(LiE[i]).TabStop := True;
    TEdit(LiE[i]).ReadOnly := False;
    TEdit(LiE[i]).Color := clWindow;
    if (Tvar[i, 0] <= 0) or (Tvar[i, 0] >= 180) then
    begin                   // Si résultat symbolique ou valeur aux limites
      TComboBox(LiB[i]).Enabled := True;
      TComboBox(LiB[i]).ItemIndex := TComboBox(LiB[i]).Items.IndexOf(Unt[i]);
      TEdit(LiE[i]).Text := '';
      TEdit(LiE[i]).Modified := False;
    end;
    Tbol[i] := TEdit(LiE[i]).Text <> '';
  end
  else
  begin
    TEdit(LiE[i]).TabStop := False;
    TEdit(LiE[i]).ReadOnly := True;
    TEdit(LiE[i]).Color := clForm;
    Tbol[i] := False;
  end;
end;

Procedure EdExit(i: Integer);
// Routine de sortie du champ de saisie d'indice i
begin
  if TEdit(LiE[i]).Modified then
  begin
    if lecs(TEdit(LiE[i]).Text, i, Tvar[i, 0]) then
      TEdit(LiE[i]).Text := ''
    else
      if mods(i) then TEdit(LiE[i]).Text := '';
    Tbol[i] := TEdit(LiE[i]).Text <> '';
    TEdit(LiE[i]).Modified := False;
  end;
end;

procedure UniChange(i: Integer);
// Routine de changement d'unité d'indice i
begin
  Unt[i] := TComboBox(LiB[i]).Items[TComboBox(LiB[i]).ItemIndex];
  TEdit(LiE[i]).Hint := chnt(i);
  if TEdit(LiE[i]).Text <> '' then
  begin
    TEdit(LiE[i]).Text := affr(modr(i), i, prc);
    TEdit(LiE[i]).Modified := False;
  end;
end;

function cas1(): Integer;
// Résolution du triangle dont on connait les 3 côtés (a, b, c)
// Solution unique si le triangle est possible
var
  pa, pb, pc, Ga, Gb, Gc: Real;  // Pour les angles en radian, appellation classique : a, b, c, A, B, C
  p: Real;
begin
  pa := DegToRad(Tvar[0, 0]); pb := DegToRad(Tvar[1, 0]); pc := DegToRad(Tvar[2, 0]);
  p := (pa + pb + pc)/2;  //Calcul du demi-périmètre
  if p >= Pi then
    Result := 1  // Le demi-périmètre d'un triangle sphérique ne peut pas dépasser Pi
  else
    if (pa >= p) or (pb >= p) or (pc >= p) then
      Result := 1 // Chaque coté d'un triangle ne peut pas dépasser le demi-périmètre (autre forme de l'inégalité triangulaire)
    else
    begin
      Ga := (Cos(pa) - Cos(pb)*Cos(pc))/Sin(pb)/Sin(pc);
      if Ga >= 1 then Ga := 0 else if Ga <= -1 then Ga := Pi else Ga := ArcCos(Ga);
      Gb := (cos(pb) - Cos(pc)*Cos(pa))/Sin(pc)/Sin(pa);
      if Gb >= 1 then Gb := 0 else if Gb <= -1 then Gb := Pi else Gb := ArcCos(Gb);
      Gc := (Cos(pc) - Cos(pa)*Cos(pb))/Sin(pa)/Sin(pb);
      if Gc >= 1 then Gc := 0 else if Gc <= -1 then Gc := Pi else Gc := ArcCos(Gc);
      Tvar[3, 0] := RadToDeg(Ga); Tvar[3, 1] := Tvar[3, 0];
      Tvar[4, 0] := RadToDeg(Gb); Tvar[4, 1] := Tvar[4, 0];
      Tvar[5, 0] := RadToDeg(Gc); Tvar[5, 1] := Tvar[5, 0];
      Result := 0;
    end;
end;

function cas3(): Integer;
// Résolution du triangle dont on connait 2 côtés et l'angle entre les côtés (a, b, C)
// Toujours une solution unique
var
  pa, pb, pc, Ga, Gb, Gc: Real;  // Pour les angles en radian, appellation classique : a, b, c, A, B, C
begin
  pa := DegToRad(Tvar[0, 0]); pb := DegToRad(Tvar[1, 0]); Gc := DegToRad(Tvar[5, 0]);
  pc := Cos(pa)*Cos(pb) + Sin(pa)*Sin(pb)*Cos(Gc);
  if pc >= 1 then
  begin
    pc := 0; Ga := Pi/2; Gb := Pi/2;
  end
  else
    if pc <= -1 then
    begin
      pc := Pi; Ga := Pi/2; Gb := Pi/2;
    end
    else
    begin
    pc := ArcCos(pc);
    Ga := (Cos(pa) - Cos(pb)*Cos(pc))/Sin(pb)/Sin(pc);
    if Ga >= 1 then Ga := 0 else if Ga <= -1 then Ga := Pi else Ga := ArcCos(Ga);
    Gb := (Cos(pb) - Cos(pc)*Cos(pa))/Sin(pc)/Sin(pa);
    if Gb >= 1 then Gb := 0 else if Gb <= -1 then Gb := Pi else Gb := ArcCos(Gb);
    end;
  Tvar[2, 0] := RadToDeg(pc); Tvar[2, 1] := Tvar[2, 0];
  Tvar[3, 0] := RadToDeg(Ga); Tvar[3, 1] := Tvar[3, 0];
  Tvar[4, 0] := RadToDeg(Gb); Tvar[4, 1] := Tvar[4, 0];
  Result := 0;
end;

function cas5(): Integer;
// Résolution du triangle dont on connait 2 côtés et l'angle à l'extrémité d'un côté (a, b, A)
// Repose sur la recherche de c dans l'équation trigonométrique E : cos a = cos b.cos c + sin b.sin c.cos A
// Nombre de solution : 0, 1, 2 ou une infinité
var
  pa, pb, pc, Ga, Gb, Gc: Real;  // Pour les angles en radian, appellation classique : a, b, c, A, B, C
  pc2, Gb2, Gc2: Real;           // Pour la seconde solution
  th: Real;
begin
  pa := DegToRad(Tvar[0, 0]); pb := DegToRad(Tvar[1, 0]); Ga := DegToRad(Tvar[3, 0]);
  if SameValue(pb, Pi/2) and SameValue(Ga, Pi/2) then
    if not SameValue(pa, Pi/2) then
      Result := 1 // L'angle pa doit être de Pi/2
    else
    begin         // E devient : 0 = 0, c peut être quelconque. C dépend de c (et vice-versa)
      pc := 3*Pi/2; Gb := Pi/2; Gc := 3*Pi/2;  // 3*Pi/2 est le codage de l'interdépendance
      Tvar[2, 0] := RadToDeg(pc); Tvar[2, 1] := Tvar[2, 0];
      Tvar[4, 0] := RadToDeg(Gb); Tvar[4, 1] := Tvar[4, 0];
      Tvar[5, 0] := RadToDeg(Gc); Tvar[5, 1] := Tvar[5, 0];
      Result := 0;
    end
  else
  begin
    th := Cos(pa)/Sqrt(Sqr(Cos(pb)) + Sqr(Sin(pb)*Cos(Ga)));
    if Abs(th) > 1 then
      Result := 1 // Pas de solution
    else
    begin
      pc := Arctan2(Sin(pb)*Cos(Ga), Cos(pb)) + Arccos(th);
      if (pc > Pi) or SameValue(pc, Pi) then pc := pc - 2*Pi;
      pc2 := Arctan2(Sin(pb)*Cos(Ga), Cos(pb)) - ArcCos(th);
      if (pc2 < -Pi) and not SameValue(pc2, -Pi) then pc2 := pc2 + 2*Pi;
      if pc < pc2 then  // Organisation des résultats en valeurs décroissantes
      begin
        th := pc; pc := pc2; pc2 := th;
      end;
      if ((pc < 0) or IsZero(pc)) and ((pc2 < 0) or IsZero(pc2)) then
      begin
        Result := 1 // Pas de solution, les résultats ne sont pas dans l'interval ]0, Pi[
      end
      else
      begin
        if (pc2 < 0) or IsZero(pc2) then pc2 := pc; // pc2 n'est pas acceptable, une seule solution
        Gb := (Cos(pb) - Cos(pc)*Cos(pa))/Sin(pc)/Sin(pa);
        if Gb >= 1 then Gb := 0 else if Gb <= -1 then Gb := Pi else Gb := ArcCos(Gb);
        Gc := (Cos(pc) - Cos(pa)*Cos(pb))/Sin(pa)/Sin(pb);
        if Gc >= 1 then Gc := 0 else if Gc <= -1 then Gc := Pi else Gc := ArcCos(Gc);
        Gb2 := (Cos(pb) - Cos(pc2)*Cos(pa))/Sin(pc2)/Sin(pa);
        if Gb2 >= 1 then Gb2 := 0 else if Gb2 <= -1 then Gb2 := Pi else Gb2 := ArcCos(Gb2);
        Gc2 := (Cos(pc2) - Cos(pa)*Cos(pb))/Sin(pa)/Sin(pb);
        if Gc2 >= 1 then Gc2 := 0 else if Gc2 <= -1 then Gc2 := Pi else Gc2 := ArcCos(Gc2);
        Tvar[2, 0] := RadToDeg(pc); Tvar[2, 1] := RadToDeg(pc2);
        Tvar[4, 0] := RadToDeg(Gb); Tvar[4, 1] := RadToDeg(Gb2);
        Tvar[5, 0] := RadToDeg(Gc); Tvar[5, 1] := RadToDeg(Gc2);
        Result := 0;
      end;
    end;
  end;
end;  

procedure invi();
// Echange des paramètres 0/1 et 3/4
var
  i: Integer;
  m0, m3: Real;
begin
  for i := 0 to 1 do
  begin
    m0 := Tvar[0, i]; m3 := Tvar[3, i];
    Tvar[0, i] := Tvar[1, i]; Tvar[3, i] := Tvar[4, i];
    Tvar[1, i] := m0; Tvar[4, i] := m3;
  end;
end;

procedure invj();
// Echange des paramètres 1/2 et 4/5
var
  i: Integer;
  m1, m4: Real;
begin
  for i := 0 to 1 do
  begin
    m1 := Tvar[1, i]; m4 := Tvar[4, i];
    Tvar[1, i] := Tvar[2, i]; Tvar[4, i] := Tvar[5, i];
    Tvar[2, i] := m1; Tvar[5, i] := m4;
  end;
end;

procedure pol();
// Passage au triangle polaire
var
  i, j: Integer;
  m: Real;
begin
  // Angles supplémentaire
  for j := 0 to 1 do
    for i := 0 to 5 do
      Tvar[i, j] := 180 - Tvar[i, j];
  // Inversion des côtés et des angles
  for j := 0 to 1 do
    for i := 0 to 2 do
    begin
      m := Tvar[i, j]; Tvar[i,j] := Tvar[i + 3, j]; Tvar[i + 3, j] := m;
    end;
end;

procedure TForm1.BtCalculClick(Sender: TObject);
var
  chx: Word;
  i, p2, nbpar, conf, cont: Integer;
  my: Real;
  merr, seq: String;
  su, sol, bl: Boolean;
begin
  //Pour quitter un éventuel champ de saisie lors de la frappe de la touche 'Entrée'
  for i := 0 to 5 do
    if ActiveControl = LiE[i] then EdExit(i);
  p2 := 1;
  nbpar := 0;
  conf := 0;
  for i := 0 to 5 do  // Calcul de la configuration
  begin
    if Tbol[i] then
    begin
      Inc(nbpar);
      conf := conf + p2;
    end;
    p2 := 2*p2;
  end;
  if (nbpar < 3) or (nbpar > 3) then  // Vérification du nombre d'angles saisis
  begin
    if nbpar < 3 then merr := 'Il faut 3 paramètres';
    if nbpar > 3 then merr := 'Seulement 3 paramètres';
    MessageDlgPos(merr, mtError, [mbOK], 0, Left, Top);
  end
  else
  begin
    for i := 0 to 5 do                // Décoche la case s'il n'y a pas de valeur saisie
      if TCheckBox(LiC[i]).Checked and not Tbol[i] then TCheckBox(LiC[i]).Checked := False;
    case conf of    // Acquisition de la séquence de traitement correspondant à la configuration
    7: seq := '1';
    11: seq := '5';
    19: seq := 'i5i';
    35: seq := '3';
    13: seq := 'j5j';
    21: seq := 'j3j';
    37: seq := 'ji5ij';
    14: seq := 'ij3ji';
    22: seq := 'ij5ji';
    38: seq := 'iji5iji';
    56: seq := 'p1p';
    25: seq := 'p5p';
    26: seq := 'pi5ip';
    28: seq := 'p3p';
    41: seq := 'pj5jp';
    42: seq := 'pj3jp';
    44: seq := 'pji5ijp';
    49: seq := 'pij3jip';
    50: seq := 'pij5jip';
    52: seq := 'piji5ijip';
    end;
    cont := 0;
    for i := 1 to Length(seq) do  // Exécution de la séquence de traitement
      case seq[i] of
      'i': invi();
      'j': invj();
      'p': begin
        pol();
        if cont > 1 then cont := cont + 10; // Pour future évolution vers un codage du message d'erreur
      end;
      '1': cont := cas1();
      '3': cont := cas3();
      '5': cont := cas5();
      end;
    if cont = 0 then
    begin                          // Pas d'erreur, affichage des résultats
      if (conf = 7) and (Grot <> 0) then //Choix du signe des angles
      begin
        chx := MessageDlgPos(affr(0, 3, 4) + ' < ' + Etiq[3] + ' < ' + affr(180, 3, 4) + ' ' + Unt[3] + ' ?',
                             mtConfirmation, [mbYes, mbNo], 0, Left, Top);
        Grot := Srot[3] div Abs(Srot[3]); if (chx = mrNo) then Grot := -Grot;
      end;
      su := True;   // A priori, solution unique
      for i := 0 to 5 do
      begin
        if conf mod 2 = 0 then
        begin               // i correspond à l'indice d'un résultat
          if (Tvar[i, 0] <> Tvar[i, 1]) and su then // Choix de la solution s'il y en a 2
          begin
            su := False;
            my := (Tvar[i, 0] + Tvar[i, 1])/2;
            if Comp[i] then my := 90 - my;     // La question est toujours posée pour un côté
            chx := MessageDlgPos('Deux solutions : ' + Etiq[i] + ' > ' + affr(my, i, 5) + ' ' + Unt[i] + ' ?',
                                  mtConfirmation, [mbYes, mbNo], 0, Left, Top);
            sol := (chx = mrNo) xor (Tvar[i, 0] < Tvar[i, 1]) xor Comp[i];  // sol est vrai si on choisit la solution Tvar[i, 1]
          end;
          if sol and not su then Tvar[i, 0] := Tvar[i, 1];
          bl := Sin(DegToRad(Tvar[i, 0])) > -0.5;
          if bl then                   // Cas des résultats "normaux"
            seq := affr(modr(i), i, prc)
          else                         // Cas des résultats symboliques
          begin
            p2 := 0;
            if i >= 3 then p2 := p2 + 1;                         // Angle
            if Comp[(i mod 3) + 3] then p2 := p2 + 2;            // Angle complémenté
            if Comp[i mod 3] then p2 := p2 + 4;                  // Côté complémenté
            if Grot*Srot[(i mod 3) + 3] = -1 then p2 := p2 + 8;  // Angle négatif
            if Grot*Srot[(i mod 3) + 3] = -2 then p2 := p2 + 16; // Angle > 180°
            case p2 of
            0, 1:   seq := Format('= %s', [Etiq[(i + 3) mod 6]]);
            2, 3:   seq := Format('= 180 - %s', [Etiq[(i + 3) mod 6]]);
            4, 5:   seq := Format('= 90 - %s', [Etiq[(i + 3) mod 6]]);
            6, 13:  seq := Format('= %s - 90', [Etiq[(i + 3) mod 6]]);
            7, 12:  seq := Format('= %s + 90', [Etiq[(i + 3) mod 6]]);
            8, 9:   seq := Format('= -%s', [Etiq[(i + 3) mod 6]]);
            10, 19: seq := Format('= %s + 180', [Etiq[(i + 3) mod 6]]);
            11, 18: seq := Format('= %s - 180', [Etiq[(i + 3) mod 6]]);
            14, 15: seq := Format('= -90 - %s', [Etiq[(i + 3) mod 6]]);
            16, 17: seq := Format('= 360 - %s', [Etiq[(i + 3) mod 6]]);
            20:     seq := Format('= %s - 270', [Etiq[(i + 3) mod 6]]);
            21:     seq := Format('= %s + 270', [Etiq[(i + 3) mod 6]]);
            22, 23: seq := Format('= 270 - %s', [Etiq[(i + 3) mod 6]]);
            end;
          end;
          TEdit(LiE[i]).Text := seq;
          TEdit(LiE[i]).Modified := False;
          TComboBox(LiB[i]).Enabled := bl;
          if bl then
            TComboBox(LiB[i]).ItemIndex := TComboBox(LiB[i]).Items.IndexOf(Unt[i])
          else
            TComboBox(LiB[i]).ItemIndex := 0;
        end;
        conf := conf div 2;
      end;
    end
    else
      MessageDlgPos('Pas de solution', mtError, [mbOK], 0, Left, Top);
  end;
end;

procedure TForm1.ChkPaClick(Sender: TObject);
begin
  ChkClick(0);
end;

procedure TForm1.ChkPbClick(Sender: TObject);
begin
  ChkClick(1);
end;

procedure TForm1.ChkPcClick(Sender: TObject);
begin
  ChkClick(2);
end;

procedure TForm1.ChkGaClick(Sender: TObject);
begin
  ChkClick(3);
end;

procedure TForm1.ChkGbClick(Sender: TObject);
begin
  ChkClick(4);
end;

procedure TForm1.ChkGcClick(Sender: TObject);
begin
  ChkClick(5);
end;

procedure TForm1.EdPaExit(Sender: TObject);
begin
  EdExit(0);
end;

procedure TForm1.EdPbExit(Sender: TObject);
begin
  EdExit(1);
end;

procedure TForm1.EdPcExit(Sender: TObject);
begin
  EdExit(2);
end;

procedure TForm1.EdGaExit(Sender: TObject);
begin
  EdExit(3);
end;

procedure TForm1.EdGbExit(Sender: TObject);
begin
  EdExit(4);
end;

procedure TForm1.EdGcExit(Sender: TObject);
begin
  EdExit(5);
end;

procedure TForm1.UniPaChange(Sender: TObject);
begin
  UniChange(0);
end;

procedure TForm1.UniPbChange(Sender: TObject);
begin
  UniChange(1);
end;

procedure TForm1.UniPcChange(Sender: TObject);
begin
  UniChange(2);
end;

procedure TForm1.UniGaChange(Sender: TObject);
begin
  UniChange(3);
end;

procedure TForm1.UniGbChange(Sender: TObject);
begin
  UniChange(4);
end;

procedure TForm1.UniGcChange(Sender: TObject);
begin
  UniChange(5);
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  fic: TextFile;
  lin: String;
  pl: Boolean;
begin
  DefaultFormatSettings.DecimalSeparator := '.';
  // Mise en place des listes pour accéder aux objets insérés par un index
  LiC := Tlist.Create;         // Liste des CheckBox
  LiC.Add(ChkPa); LiC.Add(ChkPb); LiC.Add(ChkPc); LiC.Add(ChkGa); LiC.Add(ChkGb); LiC.Add(ChkGc);
  LiE := Tlist.Create;         // Liste des Edit
  LiE.Add(EdPa); LiE.Add(EdPb); LiE.Add(EdPc); LiE.Add(EdGa); LiE.Add(EdGb); LiE.Add(EdGc);
  LiB := Tlist.Create;         // Liste des ComboBox
  LiB.Add(UniPa); LiB.Add(UniPb); LiB.Add(UniPc); LiB.Add(UniGa); LiB.Add(UniGb); LiB.Add(UniGc);
  if FileExists(GetCurrentDir + DirectorySeparator + 'TriSph.cnf') then
  begin                        // Le fichier TriSph.cnf a été trouver dans le répertoire de lancement
    AssignFile(fic, GetCurrentDir + DirectorySeparator + 'TriSph.cnf');
    Reset(fic);
    pl := True;                // On va lire la première ligne
    while not Eof(fic) do
    begin
      Readln(fic, lin);
      if pl and (UTF8Copy(lin, 2, 6) = 'TriSph') then
      begin                    // La première ligne contient une information de version
        vcf := StrToInt(UTF8Copy(lin, 8, 1));
        Readln(fic, lin);
      end;
      pl := False;             // La première ligne est lue
      Conf.Items.Add(lin);     // Ajoute le le nom de la configuration dans la boîte de sélection
      while (lin <> '---') and not Eof(fic) do
        Readln(fic, lin);
    end;
    CloseFile(fic);
  end;
  Conf.ItemIndex := 0;
  ConfChange(nil);
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
// Outil de mise au point pour afficher les valeurs calculées en pleine définition
var
  i: Integer;
begin
  if (Key = Ord('Y')) and (ssCtrl in Shift) then
  begin
    for i := 0 to 5 do                //Pour quitter un éventuel champ de saisie
      if ActiveControl = LiE[i] then EdExit(i);
    ShowMessagePos('Valeurs de Tbol[i] et Tvar[i, 0] :'
        + #13#10'0 : ' + BoolToStr(Tbol[0], True)+ ', ' + FloatToStr(Tvar[0, 0])
        + #13#10'1 : ' + BoolToStr(Tbol[1], True)+ ', ' + FloatToStr(Tvar[1, 0])
        + #13#10'2 : ' + BoolToStr(Tbol[2], True)+ ', ' + FloatToStr(Tvar[2, 0])
        + #13#10'3 : ' + BoolToStr(Tbol[3], True)+ ', ' + FloatToStr(Tvar[3, 0])
        + #13#10'4 : ' + BoolToStr(Tbol[4], True)+ ', ' + FloatToStr(Tvar[4, 0])
        + #13#10'5 : ' + BoolToStr(Tbol[5], True)+ ', ' + FloatToStr(Tvar[5, 0]),
        Left + Width, Top);
  end;
end;

procedure TForm1.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  LiC.Free;
  LiE.Free;
  LiB.Free;
  CloseAction := caFree;
end;

procedure TForm1.ConfChange(Sender: TObject);
// Changement 
var
  fic: TextFile;
  lin, mot: String;
  i,j: Integer;
begin
  // Remise à zéro de toute la fenêtre du programme
  BtRazClick(nil);
  for i := 0 to 5 do
  begin
    Tbol[i] := False;
    Etiq[i] := '';
    Comp[i] := False;
    if i >= 3 then Srot[i] := 0;
    Unt[i] := '°'
  end;
  Info.Clear;
  if Conf.ItemIndex = 0 then
  begin                  // Cas du triangle sphérique de base
    Etiq[0] := 'a';
    Etiq[1] := 'b';
    Etiq[2] := 'c';
    Etiq[3] := 'A';
    Etiq[4] := 'B';
    Etiq[5] := 'C';
    Grot := 0;
    Info.Text := 'Les minuscules sont les côtés du triangle sphérique'#13#10#13#10;
    Info.Text := Info.Text + 'Les majuscules correspondantes sont les angles opposés'#13#10;
  end
  else
  begin                  // Autres cas à lire dans TriSph.cnf
    AssignFile(fic, GetCurrentDir + DirectorySeparator + 'TriSph.cnf');
    Reset(fic);
    if vcf >= 1 then Readln(fic, lin); // Si la version n'est pas nulle, passer la première ligne
    while not Eof(fic) do // Recherche du paragraphe correspondant à la sélection
    begin
      Readln(fic, lin);
      if lin = Conf.Items[Conf.ItemIndex] then Break;
      while (lin <> '---') and not Eof(fic) do
        Readln(fic, lin);
    end;
    if Eof(fic) then
      MessageDlgPos('Ce paragraphe ne se trouve plus'#13#10'dans le fichier de configuration.'#13#10'Relancer TriSph.',
                     mtError, [mbOK], 0, Left, Top)
    else
    begin
      Readln(fic,lin); // Lecture de l'étiquette des angles
      mot := ''; j := 0;
      for i := 1 to Length(lin) + 1 do
        if (i = Length(lin) + 1) or (lin[i] = ',') then
        begin
          if j < 6 then Etiq[j] := mot;
          mot := ''; Inc(j);
        end
        else
          mot := mot + lin[i];
      Readln(fic,lin); // Lecture des booléens de complément
      for i := 1 to Length(lin) do
        if (i < 7) and (lin[i] = 'O') then Comp[i - 1] := True;
      Grot := 1;
      Readln(fic,lin); // Lecture des informations de signe
      for i := 1 to Length(lin) do
        if (i < 4) then
          case lin[i] of
          'Q': Srot[i + 2] := 2;
          'R': Srot[i + 2] := 1;
          'S': Srot[i + 2] := -1;
          'T': Srot[i + 2] := -2;
          else Grot := 0;
          end;
      if vcf >=1 then
      begin
        Readln(fic,lin); // Lecture des unités
        mot := ''; j := 0;
        for i := 1 to Length(lin) + 1 do
          if (i = Length(lin) + 1) or (lin[i] = ',') then
          begin
            if j < 6 then Unt[j] := mot;
            mot := ''; Inc(j);
          end
          else
            mot := mot + lin[i];
        Readln(fic,lin); // Lecture des valeurs
        mot := ''; j := 0;
        for i := 1 to Length(lin) + 1 do
          if (i = Length(lin) + 1) or (lin[i] = ',') then
          begin
            if (j < 6) and (Trim(mot) <> '') then
            begin
              TCheckBox(LiC[j]).Checked := True;
              TEdit(LiE[j]).Text := Trim(mot);
              TEdit(LiE[j]).Modified := True;
              EdExit(j);
            end;
            mot := ''; Inc(j);
          end
          else
            mot := mot + lin[i];
      end;
      Readln(fic,lin); // Lecture des indications
      while (lin <> '---') and not Eof(fic) do
      begin
        if lin[1] <> ';' then Info.Text := Info.Text + lin + #13#10;
        Readln(fic, lin);
      end;
      if Eof(fic) then Info.Text := Info.Text + lin + #13#10;
    end;
    CloseFile(fic);
  end;
  for i := 0 to 5 do
  begin
    TCheckBox(LiC[i]).Caption := Etiq[i];
    j := TComboBox(LiB[i]).Items.IndexOf(Unt[i]);
    if j < 0 then
    begin
      TComboBox(LiB[i]).ItemIndex := 0;
      Unt[i] := '°';
    end
    else
      TComboBox(LiB[i]).ItemIndex := j;
    TEdit(LiE[i]).Hint := chnt(i);
  end;
end;

procedure TForm1.BtRazClick(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to 5 do
  begin
    TCheckBox(LiC[i]).Checked := False;
    TComboBox(LiB[i]).Enabled := True;
    TComboBox(LiB[i]).ItemIndex := TComboBox(LiB[i]).Items.IndexOf(Unt[i]);
    TEdit(LiE[i]).Text := '';
    TEdit(LiE[i]).Modified := False;
  end;
end;

procedure TForm1.BtConfigClick(Sender: TObject);
var
  {$IFDEF WIN32}
    rt: Cardinal;
  {$ENDIF}
  {$IFDEF LINUX}
    Proc: Tprocess;
  {$ENDIF}
begin
  {$IFDEF WIN32}
    rt := ShellExecute(0, 'OPEN', Pchar('Notepad'), Pchar('TriSph.cnf'), Pchar(GetCurrentDir), SW_SHOW);
    if rt <= 32 then
    begin
      if rt = SE_ERR_FNF then
        MessageDlgPos('L''application Bloc-notes est absente', mtError, [mbOK], 0, Left, Top)
      else
        MessageDlgPos('Erreur à l''ouverture du fichier TriSph.cnf', mtError, [mbOK], 0, Left, Top);
    end;
  {$ENDIF}
  {$IFDEF LINUX}
    Proc := Tprocess.Create(nil);
    Proc.CommandLine := 'gedit ' + GetCurrentDir + DirectorySeparator + 'TriSph.cnf';
    try
      Proc.Execute;
    except
      begin
        MessageDlgPos('Erreur à la tentative d''ouvrir le fichier TriSph.cnf', mtError, [mbOK], 0, Left, Top);
        Proc.free;
      end;
    end;
  {$ENDIF}
end;

procedure TForm1.BtLisezClick(Sender: TObject);
var
  {$IFDEF WIN32}
    rt: Cardinal;
  {$ENDIF}
  {$IFDEF LINUX}
    Proc: Tprocess;
  {$ENDIF}
begin
  {$IFDEF WIN32}
    rt := ShellExecute(0, 'OPEN', Pchar('Lisezmoi.pdf'), nil, Pchar(ExtractFilePath(ParamStr(0))), SW_SHOW);
    if rt <= 32 then
    begin
      if rt = SE_ERR_FNF then
        MessageDlgPos('Le fichier Lisezmoi.pdf est absent', mtError, [mbOK], 0, Left, Top)
      else
        if rt = SE_ERR_NOASSOC then
          MessageDlgPos('Aucune application ne peut lire le fichier Lisezmoi.pdf', mtError, [mbOK], 0, Left, Top)
        else
          MessageDlgPos('Erreur à l''ouverture du fichier Lisezmoi.pdf', mtError, [mbOK], 0, Left, Top);
    end;
  {$ENDIF}
  {$IFDEF LINUX}
    if FileExists(ExtractFilePath(ParamStr(0)) + 'Lisezmoi.pdf') then
    begin
      Proc := Tprocess.Create(nil);
      Proc.CommandLine := 'evince ' + ExtractFilePath(ParamStr(0)) + 'Lisezmoi.pdf';
      try
        Proc.Execute;
      except
        begin
          MessageDlgPos('Erreur à la tentative de visualiser le fichier Lisezmoi.pdf', mtError, [mbOK], 0, Left, Top);
          Proc.free;
        end;
      end;
    end
    else
      MessageDlgPos('Le fichier Lisezmoi.pdf est absent', mtError, [mbOK], 0, Left, Top);
  {$ENDIF}
end;

end.
